# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import collections.abc
import contextlib
import dataclasses
import functools
import json
import logging
import os
import pathlib
import queue
import ssl
import threading
import traceback
import typing as t

from . import dap
from ._mp_queue import ClientMPQueue, MPProtocol, MPQueue, ServerMPQueue
from ._singleton import Singleton
from ._socket_helper import CancelledError, SocketCancellationToken

HAS_DEBUGPY = True
try:  # pragma: nocover
    import debugpy

except Exception:  # pragma: nocover
    HAS_DEBUGPY = False

log = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class PathMapping:
    """Maps a local path to the equivalent remote path."""

    local_root: str
    remote_root: str


@dataclasses.dataclass(frozen=True)
class DebugConfiguration:
    """Debug configuration data.

    Extra debug configuration data sent from the DA server after the debuggee
    has connected. This data is used to send extra data not defined in the
    debug adapter protocol needed for debug operations.

    Args:
        path_mappings: The path mappings to use when translating source paths
            with what Ansible is running with.
    """

    OUTPUT_CATEGORY: t.ClassVar[str] = "ansibug_debug_configuration"

    path_mappings: t.List[PathMapping] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True)
class PlaybookProcessInfo:
    pid: int
    address: str
    use_tls: bool
    playbook_file: t.Optional[str]

    @classmethod
    def from_json(
        self,
        data: dict[str, t.Any],
    ) -> PlaybookProcessInfo:
        return PlaybookProcessInfo(
            pid=data["pid"],
            address=data["address"],
            use_tls=data["use_tls"],
            playbook_file=data["playbook_file"],
        )

    def to_json(self) -> dict[str, t.Any]:
        return {
            "pid": self.pid,
            "address": self.address,
            "use_tls": self.use_tls,
            "playbook_file": self.playbook_file,
        }


def get_pid_info_path(pid: int) -> pathlib.Path:
    """Get the path used to store info about the ansible-playbook debug proc."""
    # It is important that changes here are also reflected in the extension code
    # so that it knows where dir and what file pattern to look for when
    # scanning for available playbooks.
    tmpdir = os.environ.get("TMPDIR", "/tmp")
    return pathlib.Path(tmpdir) / f"ansibug-pid-{pid}"


class EndStrategy(Exception):
    """Exception used to end the strategy plugin."""


class DAProtocol(MPProtocol):
    def __init__(
        self,
        debugger: AnsibleDebugger,
    ) -> None:
        self._debugger = debugger

    def on_msg_received(
        self,
        msg: dap.ProtocolMessage,
    ) -> None:
        log.info("Processing msg %r", msg)
        try:
            self._debugger.process_message(msg)
        except Exception:
            log.exception("Exception while processing msg seq %d", msg.seq)

            # The DA Server will only ever send a Request msg here, this just
            # satisfies mypy.
            req = t.cast(dap.Request, msg)
            resp = dap.ErrorResponse(
                command=req.command,
                request_seq=req.seq,
                message=f"Critical Debuggee exception received\n{traceback.format_exc()}",
            )
            self._debugger.queue_msg(resp)

    def connection_closed(
        self,
        exp: Exception | None,
    ) -> None:
        if exp:
            msg = "".join(traceback.format_exception(None, exp, exp.__traceback__))
            log.info("Socket connection failed with\n%s", msg)

        # This ensures we don't get stuck waiting for this same thread to finish.
        self._debugger.shutdown(from_send_thread=True)


class DebugState(t.Protocol):
    def continue_request(
        self,
        request: dap.ContinueRequest,
    ) -> dap.ContinueResponse:
        raise NotImplementedError()  # pragma: nocover

    def disconnect(
        self,
        request: dap.DisconnectRequest,
    ) -> None:
        raise NotImplementedError()  # pragma: nocover

    def evaluate(
        self,
        request: dap.EvaluateRequest,
    ) -> dap.EvaluateResponse:
        raise NotImplementedError()  # pragma: nocover

    def get_scopes(
        self,
        request: dap.ScopesRequest,
    ) -> dap.ScopesResponse:
        raise NotImplementedError()  # pragma: nocover

    def get_stacktrace(
        self,
        request: dap.StackTraceRequest,
    ) -> dap.StackTraceResponse:
        raise NotImplementedError()  # pragma: nocover

    def get_threads(
        self,
        request: dap.ThreadsRequest,
    ) -> dap.ThreadsResponse:
        raise NotImplementedError()  # pragma: nocover

    def get_variables(
        self,
        request: dap.VariablesRequest,
    ) -> dap.VariablesResponse:
        raise NotImplementedError()  # pragma: nocover

    def set_variable(
        self,
        request: dap.SetVariableRequest,
    ) -> dap.SetVariableResponse:
        raise NotImplementedError()  # pragma: nocover

    def step_in(
        self,
        request: dap.StepInRequest,
    ) -> None:
        raise NotImplementedError()  # pragma: nocover

    def step_out(
        self,
        request: dap.StepOutRequest,
    ) -> None:
        raise NotImplementedError()  # pragma: nocover

    def step_over(
        self,
        request: dap.NextRequest,
    ) -> None:
        raise NotImplementedError()  # pragma: nocover


@dataclasses.dataclass
class AnsibleLineBreakpoint:
    id: int
    source: dap.Source
    source_breakpoint: dap.SourceBreakpoint
    breakpoint: dap.Breakpoint
    actual_path: str
    source_path: str


class AnsibleDebugger(metaclass=Singleton):
    def __init__(self) -> None:
        self._cancel_token = SocketCancellationToken()
        self._configuration_done = threading.Event()
        self._adapter_connected = False
        self._debug_config: DebugConfiguration = DebugConfiguration()
        self._proc_pid_file = get_pid_info_path(os.getpid())
        self._proto = DAProtocol(self)
        self._send_queue: queue.Queue[dap.ProtocolMessage | None] = queue.Queue()
        self._send_thread: threading.Thread | None = None
        self._strategy_connected = threading.Condition()
        self._strategy: DebugState | None = None
        self._terminated = False

        self._thread_counter = 2  # 1 is always the main thread
        self._stackframe_counter = 1
        self._variable_counter = 1

        # Stores all the client breakpoints, key is the breakpoint number/id
        self._breakpoints: dict[int, AnsibleLineBreakpoint] = {}
        self._breakpoint_counter = 1

        # Key is the path, the value is a list of the lines in that file where:
        #   None - Line is a continuation of a breakpoint range
        #   0    - Line is not something a breakpoint can be set at.
        #   1    - Line is the start of a breakpoint range
        #
        # The lines are 1 based with the 0 index representing 0 meaning a
        # breakpoint cannot be set until the first valid entry is found. A
        # continuation means the behaviour of the previous int in the list
        # continues to apply at that line.
        #
        # Examples of None would be
        #   - block/rescue/always - cannot stop on this, bp needs to be on a
        #     task.
        #   - import_* - These tasks are seen as the imported value not as
        #     itself
        #
        # Known Problems:
        #   - import_* tasks aren't present in the Playbook block. According to
        #     these rules it will be set as a breakpoint for the previous entry
        #   - roles in a play act like import_role - same as above
        #   - always and rescue aren't seen as an exact entry, will be set to
        #     the task that preceeds it
        #   - Won't contain the remaining lines of the file - bp checks will
        #     just have to use the last entry
        self._source_info: dict[str, list[int | None]] = {}

    @contextlib.contextmanager
    def with_strategy(
        self,
        strategy: DebugState,
    ) -> collections.abc.Generator[None, None, None]:
        """Sets the current strategy.

        Sets the current strategy that is being used by Ansible. This is used
        as a way to route any requests to the strategy for it to handle
        accordingly.

        Args:
            strategy: The strategy to set.
        """
        with self._strategy_connected:
            self._strategy = strategy
            self._strategy_connected.notify_all()

        try:
            yield

        finally:
            with self._strategy_connected:
                self._strategy = None
                self._strategy_connected.notify_all()

    def next_thread_id(self) -> int:
        tid = self._thread_counter
        self._thread_counter += 1

        return tid

    def next_stackframe_id(self) -> int:
        sfid = self._stackframe_counter
        self._stackframe_counter += 1

        return sfid

    def next_variable_id(self) -> int:
        vid = self._variable_counter
        self._variable_counter += 1

        return vid

    def wait_for_config_done(
        self,
        timeout: float | None,
    ) -> None:
        """Waits until the debug config is done.

        Waits until the client has sent through the configuration done request
        that indicates no more initial configuration data is expected.

        Args:
            timeout: The maximum time, in seconds, to wait until the debug
                adapter is connected and ready.
        """
        self._configuration_done.wait(timeout=timeout)

    def get_breakpoint(
        self,
        path: str,
        line: int,
    ) -> AnsibleLineBreakpoint | None:
        """Gets the breakpoint for the path/line specified.

        Checks to see if there is a breakpoint set for the path and line that
        is requested. If no breakpoint is associated with the arguments then
        None will be returned.

        Args:
            path: The path of the file to check.
            line: The line in the file to check.

        Returns:
            Optional[AnsibleLineBreakpoint]: The breakpoint associated with the
            args if present.
        """
        for b in self._breakpoints.values():
            if (
                b.actual_path == path
                and (b.breakpoint.line is None or b.breakpoint.line <= line)
                and (b.breakpoint.end_line is None or b.breakpoint.end_line >= line)
            ):
                return b

        return None

    def start(
        self,
        addr: str,
        mode: t.Literal["connect", "listen"],
        *,
        ssl_context: ssl.SSLContext | None = None,
        playbook_file: str | None = None,
    ) -> str:
        """Start the background server thread.

        Starts the background server thread which starts the debuggee socket
        with the debug adapter server. This will also queue up the
        InitializedEvent to be sent to the debug adapter on connection.

        Args:
            addr: The socket address to connect/bind.
            mode: The socket mode to use, connect will connect to the addr
                while listen will bind to the addr and wait for a connection.
            ssl_context: Optional client SSLContext to wrap the socket
                connection with.
            playbook_file: The file of the playbook being run, this is optional
                info used for storing metadata with the process pid on listen.

        returns:
            str: The socket address that is being used, this is only valid when
            using listen mode.
        """
        log.info("Setting up ansible-playbook debug %s socket at '%s'", mode, addr)

        queue_kwargs: dict[str, t.Any] = {
            "address": addr,
            "proto": lambda: self._proto,
            "ssl_context": ssl_context,
            "cancel_token": self._cancel_token,
        }

        mp_queue: MPQueue
        if mode == "connect":
            mp_queue = ClientMPQueue(**queue_kwargs)
            addr = mp_queue.address

        else:
            mp_queue = ServerMPQueue(**queue_kwargs)
            addr = mp_queue.address

            proc_info = PlaybookProcessInfo(
                pid=os.getpid(),
                address=addr,
                use_tls=ssl_context is not None,
                playbook_file=playbook_file,
            )
            proc_json = json.dumps(proc_info.to_json())
            self._proc_pid_file.write_text(proc_json)

        log.debug("MPQueue set on address %s", addr)

        self._send_thread = threading.Thread(
            target=self._send_task,
            args=(mp_queue,),
            name="ansibug-debugger",
        )
        self._send_thread.start()

        return addr

    def shutdown(
        self,
        from_send_thread: bool = False,
    ) -> None:
        """Shutdown the Debuggee.

        Marks the debuggee as completed and signals the send thread to
        shutdown.

        Args:
            from_send_thread: The shutdown request has come from the send
                thread.
        """
        log.debug("Shutting down DebugServer")

        if not from_send_thread and self._send_thread:
            # If we aren't connected we need to cancel the queue start method
            # before trying to join the thread. If we are connected we want to
            # add None to the queue for the thread to exit gracefully once all
            # messages have been sent.
            if not self._adapter_connected:
                self._cancel_token.cancel()

            self.queue_msg(None)
            self._send_thread.join()
            self._send_thread = None

        # If the strategy is still running we don't want it to fire on any
        # breakpoints and our debug config no longer applies
        self._debug_config = DebugConfiguration()
        self._breakpoints = {}

        # Ensure the callback plugin isn't stuck waiting for this
        self._configuration_done.set()

        # Ensure the strategy isn't waiting for a response to anything
        with self._strategy_connected:
            if self._strategy and not self._terminated:
                self._strategy.disconnect(
                    dap.DisconnectRequest(
                        restart=False,
                        terminate_debuggee=False,
                        suspend_debuggee=False,
                    )
                )

        self._proc_pid_file.unlink(missing_ok=True)

        log.debug("DebugServer is shutdown")

    def queue_msg(
        self,
        msg: dap.ProtocolMessage | None,
    ) -> None:
        """Queues a message to send to the debug adapter.

        Adds the message to the send queue for the send thread to send to the
        debug adapter.

        Args:
            msg: The message to send.
        """
        self._send_queue.put(msg)

    def convert_to_client_path(
        self,
        path: str,
    ) -> str:
        """Converts the path to the client path.

        Converts the path provided to the client path equivalent based on the
        mappings provided. This is used in cases where the paths in the client
        does not match up with what Ansible is running with, for example
        debugging on a different host with a different path root.

        Args:
            path: The path to convert.

        Returns:
            str: The converted path.
        """
        for mapping in self._debug_config.path_mappings:
            if path.startswith(mapping.remote_root):
                return mapping.local_root + path[len(mapping.remote_root) :]

        else:
            return path

    def register_path_breakpoint(
        self,
        path: str,
        line: int,
        bp_type: int,
    ) -> None:
        """Register a valid breakpoint section.

        Registers a line as a valid breakpoint in a path. This registration is
        used when responding the breakpoint requests from the client.

        Args:
            path: The file path the line is set.
            line: The line the breakpoint is registered to.
            bp_type: Set to 1 for a valid breakpoint and 0 for an invalid
                breakpoint section.
        """
        # Ensure each new entry has a starting value of 0 which denotes that
        # a breakpoint cannot be set at the start of the file. It can only be
        # set when a line was registered.
        file_lines = self._source_info.setdefault(path, [0])
        file_lines.extend([None] * (1 + line - len(file_lines)))
        file_lines[line] = bp_type

        for breakpoint in self._breakpoints.values():
            if breakpoint.actual_path != path:
                continue

            source_breakpoint = breakpoint.source_breakpoint
            bp = self._calculate_breakpoint(
                breakpoint.id,
                breakpoint.source,
                source_breakpoint.line,
                file_lines,
            )

            if (
                breakpoint.breakpoint.verified != bp.verified
                or breakpoint.breakpoint.line != bp.line
                or breakpoint.breakpoint.end_line != bp.end_line
            ):
                breakpoint.breakpoint = bp
                self.queue_msg(
                    dap.BreakpointEvent(
                        reason="changed",
                        breakpoint=bp,
                    )
                )

    @classmethod
    def _enable_debugpy(cls) -> None:  # pragma: nocover
        """This is only meant for debugging ansibug in Ansible purposes."""
        if not HAS_DEBUGPY:
            raise Exception("Failed to enable debugging because debugpy is not installed")

        elif not debugpy.is_client_connected():
            debugpy.listen(("localhost", 12535))
            debugpy.wait_for_client()

    def _get_strategy(self) -> DebugState:
        with self._strategy_connected:
            self._strategy_connected.wait_for(lambda: self._strategy is not None)
            return t.cast(DebugState, self._strategy)

    @functools.singledispatchmethod
    def process_message(
        self,
        msg: dap.ProtocolMessage,
    ) -> None:
        raise NotImplementedError(f"Debuggee does not support the {type(msg).__name__} message")

    @process_message.register
    def _(
        self,
        msg: dap.ConfigurationDoneRequest,
    ) -> None:
        resp = dap.ConfigurationDoneResponse(request_seq=msg.seq)
        self.queue_msg(resp)
        self._configuration_done.set()

    @process_message.register
    def _(
        self,
        msg: dap.ContinueRequest,
    ) -> None:
        strategy = self._get_strategy()
        resp = strategy.continue_request(msg)
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.DisconnectRequest,
    ) -> None:
        self._disconnect(msg)
        # The response is sent by the debug adapter.

    @process_message.register
    def _(
        self,
        msg: dap.EvaluateRequest,
    ) -> None:
        strategy = self._get_strategy()
        resp = strategy.evaluate(msg)
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.NextRequest,
    ) -> None:
        strategy = self._get_strategy()
        strategy.step_over(msg)
        self.queue_msg(dap.NextResponse(request_seq=msg.seq))

    @process_message.register
    def _(
        self,
        msg: dap.ScopesRequest,
    ) -> None:
        strategy = self._get_strategy()
        resp = strategy.get_scopes(msg)
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.SetBreakpointsRequest,
    ) -> None:
        # FIXME: Deal with source_reference if set
        source_path = msg.source.path or ""

        # If explicit path mappings are provided we convert the client provided
        # source path to the path known to Ansible. This is important as the
        # source mappings are keyed by the Ansible paths.
        actual_path = source_path
        for mapping in self._debug_config.path_mappings:
            if actual_path.startswith(mapping.local_root):
                actual_path = mapping.remote_root + actual_path[len(mapping.local_root) :]
                break

        source_info = self._source_info.get(actual_path, None)

        # Clear out existing breakpoints for the source as each request should
        # send the latest list for a source.
        self._breakpoints = {bpid: b for bpid, b in self._breakpoints.items() if b.source_path != source_path}

        breakpoint_info: list[dap.Breakpoint] = []
        for source_breakpoint in msg.breakpoints:
            bp_id = self._breakpoint_counter
            self._breakpoint_counter += 1

            bp: dap.Breakpoint
            if msg.source_modified:
                bp = dap.Breakpoint(
                    id=bp_id,
                    verified=False,
                    message="Cannot set breakpoint on a modified source.",
                    source=msg.source,
                )
                # I don't think we need to preserve this bp for later reference.
                breakpoint_info.append(bp)
                continue

            if not source_info:
                bp = dap.Breakpoint(
                    id=bp_id,
                    verified=False,
                    message="File has not been loaded by Ansible, cannot detect breakpoints yet.",
                    source=msg.source,
                    line=source_breakpoint.line,
                )

            else:
                bp = self._calculate_breakpoint(
                    bp_id,
                    msg.source,
                    source_breakpoint.line,
                    source_info,
                )

            self._breakpoints[bp_id] = AnsibleLineBreakpoint(
                id=bp_id,
                source=msg.source,
                source_breakpoint=source_breakpoint,
                breakpoint=bp,
                actual_path=actual_path,
                source_path=source_path,
            )
            breakpoint_info.append(bp)

        resp = dap.SetBreakpointsResponse(
            request_seq=msg.seq,
            breakpoints=breakpoint_info,
        )
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.SetVariableRequest,
    ) -> None:
        strategy = self._get_strategy()
        resp = strategy.set_variable(msg)
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.StackTraceRequest,
    ) -> None:
        strategy = self._get_strategy()
        resp = strategy.get_stacktrace(msg)
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.StepInRequest,
    ) -> None:
        strategy = self._get_strategy()
        strategy.step_in(msg)
        self.queue_msg(dap.StepInResponse(request_seq=msg.seq))

    @process_message.register
    def _(
        self,
        msg: dap.StepOutRequest,
    ) -> None:
        strategy = self._get_strategy()
        strategy.step_out(msg)
        self.queue_msg(dap.StepOutResponse(request_seq=msg.seq))

    @process_message.register
    def _(
        self,
        msg: dap.TerminateRequest,
    ) -> None:
        self._terminated = True
        self._disconnect(dap.DisconnectRequest(terminate_debuggee=True))
        self.queue_msg(dap.TerminateResponse(request_seq=msg.seq))

    @process_message.register
    def _(
        self,
        msg: dap.ThreadsRequest,
    ) -> None:
        strategy = self._get_strategy()
        resp = strategy.get_threads(msg)
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.VariablesRequest,
    ) -> None:
        strategy = self._get_strategy()
        resp = strategy.get_variables(msg)
        self.queue_msg(resp)

    @process_message.register
    def _(
        self,
        msg: dap.OutputEvent,
    ) -> None:
        # This should always be this category as only the da server sends this
        # message to the debuggee.
        if msg.category != DebugConfiguration.OUTPUT_CATEGORY or not isinstance(
            msg.data, DebugConfiguration
        ):  # pragma: nocover
            return

        self._debug_config = msg.data

        # Wait until the debug adapter has sent this message before sending the
        # InitializedEvent to the client.
        self.queue_msg(dap.InitializedEvent())

    def _calculate_breakpoint(
        self,
        bp_id: int,
        bp_source: dap.Source,
        line: int,
        file_lines: list[int | None],
    ) -> dap.Breakpoint:
        """Builds a Breakpoint object based on the source provided and known file_lines."""
        start_line = min(line, len(file_lines) - 1)
        end_line = start_line + 1

        line_type = file_lines[start_line]
        while line_type is None:
            start_line -= 1
            line_type = file_lines[start_line]

        while end_line < len(file_lines) and file_lines[end_line] is None:
            end_line += 1

        end_line = min(end_line - 1, len(file_lines))

        if line_type == 0:
            verified = False
            bp_msg = "Breakpoint cannot be set here."
        else:
            verified = True
            bp_msg = None

        return dap.Breakpoint(
            id=bp_id,
            verified=verified,
            message=bp_msg,
            source=bp_source,
            line=start_line,
            end_line=end_line,
        )

    def _disconnect(
        self,
        msg: dap.DisconnectRequest,
    ) -> None:
        self._debug_config = DebugConfiguration()
        self._breakpoints = {}

        strategy = self._get_strategy()
        strategy.disconnect(msg)
        self._send_queue.put(None)

    def _send_task(
        self,
        mp_queue: MPQueue,
    ) -> None:
        """Debuggee send task

        This is the task that waits for the debug adapter to be connected to
        the socket and continuously sends the queued messages back to the
        adapter.

        Args:
            mp_queue: The queue to use for communication.
        """
        log.debug("Debuggee server send thread started")
        try:
            mp_queue.start()
            log.debug("Debuggee socket is connected and ready for use.")
            self._adapter_connected = True

            while msg := self._send_queue.get():
                log.info("Sending to debug adapter %r", msg)
                mp_queue.send(msg)

        except CancelledError:
            pass

        except Exception:
            log.exception("Unknown error in Debuggee send thread")

        finally:
            mp_queue.stop()
            self._adapter_connected = False

        log.debug("Debuggee server send thread ended")



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
name: debug
short_description: Executes tasks with a Debug Adapter Protocol debugger
description:
- Acts like the linear strategy plugin but adds functionality for interacting
  with a Debug Adapter Protocol (DAP) debugger, like one used by
  Visual Studio Code.
author: Jordan Borean (@jborean93)
"""

import collections.abc
import enum
import inspect
import logging
import os
import threading
import traceback
import typing as t

from ansible import constants as C
from ansible.errors import AnsibleError, AnsibleUndefinedVariable
from ansible.executor.play_iterator import PlayIterator
from ansible.executor.task_queue_manager import TaskQueueManager
from ansible.executor.task_result import TaskResult
from ansible.inventory.host import Host
from ansible.parsing.dataloader import DataLoader
from ansible.playbook.block import Block
from ansible.playbook.conditional import Conditional
from ansible.playbook.included_file import IncludedFile
from ansible.playbook.play_context import PlayContext
from ansible.playbook.task import Task
from ansible.plugins.strategy.linear import StrategyModule as LinearStrategy
from ansible.template import AnsibleNativeEnvironment, Templar
from ansible.utils.display import Display
from ansible.vars.manager import VariableManager

import ansibug
from ansibug._debuggee import (
    AnsibleDebugger,
    AnsibleLineBreakpoint,
    DebugState,
    EndStrategy,
)

from ..plugin_utils._breakpoints import register_block_breakpoints
from ..plugin_utils._repl_util import (
    RemoveVarCommand,
    SetVarCommand,
    TemplateCommand,
    parse_repl_args,
)

display = Display()
log = logging.getLogger("ansibug.strategy")


class ThreadState(enum.Enum):
    CONTINUE = enum.auto()
    """Thread will continue to run until a breakpoint is hit."""

    STEP_IN = enum.auto()
    """Thread should step in and break on the next task."""

    STEP_OUT = enum.auto()
    """Thread should step out and break on the next task outside of the stack."""

    STEP_OVER = enum.auto()
    """Thread should step over and break on the next task in the same stack."""

    WAIT = enum.auto()
    """Thread has hit a breakpoint and is waiting for the client to respond."""

    END = enum.auto()
    """Thread has been requested to end and should not process any more tasks."""


class AnsibleThread:
    def __init__(
        self,
        *,
        id: int,
        host: Host | None = None,
    ) -> None:
        self.id = id
        self.host = host
        self.stack_frames: list[int] = []
        self.state: ThreadState = ThreadState.CONTINUE
        self._stopped_task: Task | None = None

    def to_dap(self) -> ansibug.dap.Thread:
        return ansibug.dap.Thread(
            id=self.id,
            name=self.host.get_name() if self.host else "main",
        )

    def set_state(
        self,
        state: ThreadState,
    ) -> None:
        log.debug("Setting thread %d state to %s->%s", self.id, self.state, state)

        if state == ThreadState.STEP_OUT and self._stopped_task:
            # If a thread has been set to step out we want to make comparisons
            # later on easier by setting the stopped task to the parent include
            # task.
            task = self._stopped_task
            while task := task._parent:
                if isinstance(task, Task) and task.action in C._ACTION_ALL_INCLUDES:
                    break

            self._stopped_task = task

        elif (
            state == ThreadState.STEP_IN
            and self._stopped_task
            and self._stopped_task.action not in C._ACTION_ALL_INCLUDES
        ):
            # If changing to STEP_IN but the stopped task was not an include,
            # treat it like STEP_OVER.
            state = ThreadState.STEP_OVER

        self.state = state

    def should_break(
        self,
        task: Task,
        breakpoint: AnsibleLineBreakpoint | None,
    ) -> ansibug.dap.StoppedEvent | None:
        stopped_kwargs: dict[str, t.Any] = {}

        if self._break_step_over(task):
            stopped_kwargs = {
                "reason": ansibug.dap.StoppedReason.STEP,
                "description": "Step over",
                "thread_id": self.id,
            }
            self._stopped_task = task

        elif self._break_step_out(task):
            stopped_kwargs = {
                "reason": ansibug.dap.StoppedReason.STEP,
                "description": "Step out",
                "thread_id": self.id,
            }

        elif self._break_step_in():
            stopped_kwargs = {
                "reason": ansibug.dap.StoppedReason.STEP,
                "description": "Step in",
                "thread_id": self.id,
            }

        elif breakpoint and self.state != ThreadState.STEP_OUT:
            # Breakpoints are ignored when stepping out.
            stopped_kwargs = {
                "reason": ansibug.dap.StoppedReason.BREAKPOINT,
                "description": "Breakpoint hit",
                "hit_breakpoint_ids": [breakpoint.id],
                "thread_id": self.id,
            }

        if stopped_kwargs:
            # Store this task so the step_* actions can use it when verifying
            # if the requested action for this task should cause a stop.
            self._stopped_task = task
            self.state = ThreadState.WAIT
            return ansibug.dap.StoppedEvent(**stopped_kwargs)

        return None

    def _break_step_over(
        self,
        task: Task,
    ) -> bool:
        if self.state != ThreadState.STEP_OVER or not self._stopped_task:
            return False

        # Get the parent task of the current task being evaluated as well as
        # the task which was resumed with step over. If they are the same then
        # the task should stop.
        task_parent = self._get_parent_task(task)
        stopped_parent = self._get_parent_task(self._stopped_task)

        return getattr(task_parent, "_uuid", None) == getattr(stopped_parent, "_uuid", None)

    def _break_step_in(self) -> bool:
        # If in, then the first task to call this will need to break.
        return self.state == ThreadState.STEP_IN

    def _break_step_out(
        self,
        task: Task,
    ) -> bool:
        if self.state != ThreadState.STEP_OUT or not self._stopped_task:
            return False

        # If step out, we only want to break if the current task's parent does
        # not match the stopped task stored at state change (the include). If
        # the parent matches it means we are still in the same include scope.
        task_parent = self._get_parent_task(task)
        return getattr(task_parent, "_uuid", None) != self._stopped_task._uuid

    def _get_parent_task(
        self,
        task: Task,
    ) -> Task | None:
        while task := task._parent:
            if isinstance(task, Task):
                return task

        return None


class AnsibleStackFrame:
    def __init__(
        self,
        *,
        id: int,
        task: Task,
        task_vars: dict[str, t.Any],
        debugger: AnsibleDebugger,
    ) -> None:
        self.id = id
        self.task = task
        self.task_vars = task_vars
        self.scopes: list[int] = []
        self.variables: list[int] = []
        self.variables_options_id = 0
        self.variables_taskvar_id = 0
        self.variables_hostvars_id = 0

        self._debugger = debugger

    def to_dap(self) -> ansibug.dap.StackFrame:
        task_path = self.task.get_path()

        task_path_and_line = task_path.rsplit(":", 1)
        path = task_path_and_line[0]
        line = int(task_path_and_line[1])
        client_path = self._debugger.convert_to_client_path(path)
        source = ansibug.dap.Source(name=os.path.basename(client_path), path=client_path)

        return ansibug.dap.StackFrame(
            id=self.id,
            name=self.task.get_name(),
            source=source,
            line=line,
        )


class AnsibleVariable:
    """Structure needed for an AnsibleVariable implementation."""

    def __init__(
        self,
        id: int,
        stackframe: AnsibleStackFrame,
    ) -> None:
        self.id = id
        self.stackframe = stackframe

    @property
    def named_variables(self) -> int:
        return 0

    @property
    def indexed_variables(self) -> int:
        return 0

    def get(self) -> collections.abc.Iterable[tuple[str, t.Any]]:
        raise NotImplementedError()  # pragma: nocover

    def remove(
        self,
        index: str,
    ) -> None:
        raise NotImplementedError()  # pragma: nocover

    def set(
        self,
        index: str,
        value: t.Any,
    ) -> None:
        raise NotImplementedError()  # pragma: nocover


class AnsibleListVariable(AnsibleVariable):
    """AnsibleVariable with a list datastore."""

    def __init__(
        self,
        id: int,
        stackframe: AnsibleStackFrame,
        value: collections.abc.Sequence[t.Any],
    ) -> None:
        super().__init__(id, stackframe)
        self._ds = value

    @property
    def indexed_variables(self) -> int:
        return len(self._ds)

    def get(self) -> collections.abc.Iterable[tuple[str, t.Any]]:
        return iter((str(k), v) for k, v in enumerate(self._ds))

    def set(
        self,
        index: str,
        value: t.Any,
    ) -> None:
        self._ds[int(index)] = value  # type: ignore[index]


class AnsibleDictVariable(AnsibleVariable):
    """AnsibleVariable with a dict datastore."""

    def __init__(
        self,
        id: int,
        stackframe: AnsibleStackFrame,
        value: collections.abc.Mapping[t.Any, t.Any],
    ) -> None:
        super().__init__(id, stackframe)
        self._ds = value

    @property
    def named_variables(self) -> int:
        return len(self._ds)

    def get(self) -> collections.abc.Iterable[tuple[str, t.Any]]:
        return iter((str(k), v) for k, v in self._ds.items())

    def remove(
        self,
        index: str,
    ) -> None:
        if index in self._ds:
            del self._ds[index]  # type: ignore[attr-defined]

    def set(
        self,
        index: str,
        value: t.Any,
    ) -> None:
        self._ds[index] = value  # type: ignore[index]


class AnsibleDictWithRawStoreVariable(AnsibleDictVariable):
    """AnsibleDictVariable with a secondary/raw datastore to replicate changes to."""

    def __init__(
        self,
        id: int,
        stackframe: AnsibleStackFrame,
        value: collections.abc.Mapping[t.Any, t.Any],
        raw: dict[t.Any, t.Any],
    ) -> None:
        super().__init__(id, stackframe, value)
        self._raw_ds = raw

    def remove(self, index: str) -> None:
        super().remove(index)
        self._raw_ds.pop(index, None)

    def set(self, index: str, value: t.Any) -> None:
        super().set(index, value)
        self._raw_ds[index] = value


class AnsibleHostVarsVariable(AnsibleDictVariable):
    """AnsibleDictVariable with integration into a host variable manager."""

    def __init__(
        self,
        id: int,
        stackframe: AnsibleStackFrame,
        value: collections.abc.Mapping[t.Any, t.Any],
        host: str,
        variable_manager: VariableManager,
    ) -> None:
        super().__init__(id, stackframe, value)
        self._host = host
        self._variable_manager = variable_manager

    def set(self, index: str, value: t.Any) -> None:
        super().set(index, value)
        # Persisting hostvars need to be done as a host_variable.
        self._variable_manager.set_host_variable(self._host, index, value)


class AnsibleDebugState(DebugState):
    """Ansible Debug State.

    A class used to handle the debug state with a strategy plugin. It exposes
    the methods needed for interaction with the ansibug debugger.

    Args:
        debugger: The debugger instance connected to the debug adapter.
        loader: The data loader used by the strategy plugin.
        variable_manager: The variable manager for the strategy.

    Attributes:
        threads: The debug adapter threads.
        stackframes: The debug adapter stack frames.
        variables: The debug adapter variables.
    """

    def __init__(
        self,
        debugger: AnsibleDebugger,
        loader: DataLoader,
        variable_manager: VariableManager,
    ) -> None:
        self.threads: dict[int, AnsibleThread] = {1: AnsibleThread(id=1, host=None)}
        self.stackframes: dict[int, AnsibleStackFrame] = {}
        self.variables: dict[int, AnsibleVariable] = {}

        self._debugger = debugger
        self._loader = loader
        self._variable_mamanger = variable_manager
        self._waiting_condition = threading.Condition()

    def start_task(
        self,
        host: Host,
        task: Task,
        task_vars: dict[str, t.Any],
    ) -> None:
        """Start a task.

        Processes the Ansible task before it is run. Processing a task will
        synchronize the stack frames by adding or removing the frames needed
        as well as check if a stopped even should be triggered.

        Args:
            host: The inventory host for the task.
            task: The task to process.
            task_vars: The task variables.
        """
        thread: AnsibleThread | None = next(
            iter([t for t in self.threads.values() if t.host == host]),
            None,
        )
        if not thread:
            thread = self._add_thread(host)

        if thread.stack_frames:
            # The parent is the implicit block and we want the parent of that.
            parent_task = task._parent._parent
            last_frame_id = thread.stack_frames[0]
            last_frame = self.stackframes[last_frame_id]
            if not parent_task or (last_frame.task and last_frame.task._uuid != parent_task._uuid):
                thread.stack_frames.pop(0)
                self.stackframes.pop(last_frame_id)
                for variable_id in last_frame.variables:
                    del self.variables[variable_id]

            # If this is the first task in a role included by include_role we
            # need to scan the tasks and handlers to validate the breakpoints
            if (
                parent_task
                and last_frame.task
                and last_frame.task.action in C._ACTION_INCLUDE_ROLE
                and hasattr(task, "_role")
            ):
                task_handlers = task._role.get_task_blocks()
                role_handlers = task._role.get_handler_blocks(task.play)
                register_block_breakpoints(self._debugger, task_handlers)
                register_block_breakpoints(self._debugger, role_handlers)

        sfid = self._debugger.next_stackframe_id()

        sf = self.stackframes[sfid] = AnsibleStackFrame(
            id=sfid,
            task=task,
            task_vars=task_vars,
            debugger=self._debugger,
        )
        thread.stack_frames.insert(0, sfid)

        # Some implicit meta tasks are added without a loader so won't have
        # a path. We can't set breakpoints on these tasks so skip the check.
        task_path = task.get_path()
        if not task_path:
            return

        path_and_line = task_path.rsplit(":", 1)
        path = path_and_line[0]
        line = int(path_and_line[1])

        templar = Templar(loader=self._loader, variables=sf.task_vars)

        with self._waiting_condition:
            tid = thread.id

            line_breakpoint = self._debugger.get_breakpoint(path, line)
            if line_breakpoint and line_breakpoint.source_breakpoint.condition:
                cond = Conditional(loader=self._loader)
                cond.when = [line_breakpoint.source_breakpoint.condition]

                try:
                    if not cond.evaluate_conditional(templar, templar.available_variables):
                        line_breakpoint = None
                except AnsibleError:
                    # Treat a broken template as a false condition result.
                    line_breakpoint = None

            stopped_event = thread.should_break(task, line_breakpoint)

            if stopped_event:
                self._debugger.queue_msg(stopped_event)
                self._waiting_condition.wait_for(lambda: self.threads[tid].state != ThreadState.WAIT)

                # If thread is marked as end then the client has requested us
                # to terminate the debuggee.
                if thread.state == ThreadState.END:
                    raise EndStrategy()

    def end_task(
        self,
        host: Host,
        task: Task,
    ) -> None:
        """Ends a task.

        Marks the task on the host specified as complete adjusting the stack
        frames for that host as needed.

        Args:
            host: The inventory host for the task.
            task: The task to process.
        """
        thread = next(iter([t for t in self.threads.values() if t.host == host]))

        if task.action not in C._ACTION_ALL_INCLUDES:
            sfid = thread.stack_frames.pop(0)
            sf = self.stackframes.pop(sfid)
            for variable_id in sf.variables:
                del self.variables[variable_id]

    def exit_threads(self) -> None:
        """Exits all active threads.

        Exits all active threads and passes along the exit event to the debug
        adapter. This is used by the strategy plugin when the inventory has
        been refreshed or the strategy is complete.
        """
        for tid in list(self.threads.keys()):
            if tid == 1:
                continue

            del self.threads[tid]
            self._debugger.queue_msg(
                ansibug.dap.ThreadEvent(
                    reason="exited",
                    thread_id=tid,
                )
            )

    def continue_request(
        self,
        request: ansibug.dap.ContinueRequest,
    ) -> ansibug.dap.ContinueResponse:
        if request.single_thread:
            self._resume_threads([request.thread_id], ThreadState.CONTINUE)
            all_threads_continued = False
        else:
            self._resume_threads(self.threads.keys(), ThreadState.CONTINUE)
            all_threads_continued = True

        return ansibug.dap.ContinueResponse(
            request_seq=request.seq,
            all_threads_continued=all_threads_continued,
        )

    def disconnect(
        self,
        request: ansibug.dap.DisconnectRequest,
    ) -> None:
        state = ThreadState.END if request.terminate_debuggee else ThreadState.CONTINUE
        self._resume_threads(list(self.threads.keys()), state)

    def get_scopes(
        self,
        request: ansibug.dap.ScopesRequest,
    ) -> ansibug.dap.ScopesResponse:
        sf = self.stackframes[request.frame_id]

        # This is a very basic templating of the args and doesn't handle loops.
        templar = Templar(loader=self._loader, variables=sf.task_vars)
        task_args = templar.template(sf.task.args, fail_on_undefined=False)
        module_opts = self._add_variable(
            sf,
            task_args,
            # Any changes to our templated var also needs to reflect back onto
            # the raw datastore so use a custom variable class.
            var_factory=lambda i: AnsibleDictWithRawStoreVariable(i, sf, task_args, sf.task.args),
        )
        sf.variables_options_id = module_opts.id

        task_vars = self._add_variable(sf, sf.task_vars)
        sf.variables_taskvar_id = task_vars.id

        # We use the hostvars but for simplicity sake we also overlay the task
        # vars that might be set as these will contain things like play/task
        # vars. The hostvars are a more persistent set of vars that last beyond
        # this task so is important to give the user a way to set these
        # persistently in the debugger.
        host_vars_amalgamated = {k: v for k, v in sf.task_vars["hostvars"][sf.task_vars["inventory_hostname"]].items()}
        for k, v in task_vars.get():
            if k not in host_vars_amalgamated:
                host_vars_amalgamated[k] = v

        host_vars = self._add_variable(
            sf,
            host_vars_amalgamated,
            var_factory=lambda i: AnsibleHostVarsVariable(
                i, sf, host_vars_amalgamated, sf.task_vars["inventory_hostname"], self._variable_mamanger
            ),
        )
        sf.variables_hostvars_id = host_vars.id

        global_vars = self._add_variable(sf, sf.task_vars["vars"])

        scopes: list[ansibug.dap.Scope] = [
            # Options for the module itself
            ansibug.dap.Scope(
                name="Module Options",
                variables_reference=module_opts.id,
                named_variables=module_opts.named_variables,
                indexed_variables=module_opts.indexed_variables,
            ),
            # Variables sent to the worker, complete snapshot for the task.
            ansibug.dap.Scope(
                name="Task Variables",
                variables_reference=task_vars.id,
                named_variables=task_vars.named_variables,
                indexed_variables=task_vars.indexed_variables,
            ),
            # Scoped task vars but for the current host
            ansibug.dap.Scope(
                name="Host Variables",
                variables_reference=host_vars.id,
                named_variables=host_vars.named_variables,
                indexed_variables=host_vars.indexed_variables,
                expensive=True,
            ),
            # Scoped task vars but the full vars dict
            ansibug.dap.Scope(
                name="Global Variables",
                variables_reference=global_vars.id,
                named_variables=global_vars.named_variables,
                indexed_variables=global_vars.indexed_variables,
                expensive=True,
            ),
        ]

        return ansibug.dap.ScopesResponse(
            request_seq=request.seq,
            scopes=scopes,
        )

    def evaluate(
        self,
        request: ansibug.dap.EvaluateRequest,
    ) -> ansibug.dap.EvaluateResponse:
        value = ""
        value_type = None

        # Known contexts and how they are used in VSCode
        # repl - Debug Console with the expression entered
        # watch - WATCH pane with an expression set
        # clipboard - VARIABLES pane, right click variable -> 'Copy Value'
        # variables - same as clipboard but only present for older code
        # hover - not used as we don't set the capability, a bit dangerous to enable IMO

        expression = request.expression

        if request.context == "repl" and expression.startswith("!") and request.frame_id:
            # The following repl commands are available, all start with !
            # ro - remove option
            #   Removes the option specified from the current module options
            # so - set option
            #   Add/sets the option on the module options.
            # sh - set hostvar
            #   Adds/sets the variable on the host vars.
            # t - template
            #   Templates the expression (default behaviour without !)
            # The argparse library is used to parse this data rather than do
            # it manually.
            sf = self.stackframes[request.frame_id]

            repl_command = parse_repl_args(expression[1:])
            if isinstance(repl_command, TemplateCommand):
                value, value_type = self._safe_evaluate_expression(repl_command.expression, sf.task_vars)

            elif isinstance(repl_command, RemoveVarCommand):
                ansible_var = self.variables[sf.variables_options_id]
                ansible_var.remove(repl_command.name)

            elif isinstance(repl_command, SetVarCommand):
                new_value = self._template(repl_command.expression, sf.task_vars)

                var_ids = []
                if repl_command.command == "set_option":
                    var_ids += [sf.variables_options_id]
                else:
                    # We want to have changing hostvars also apply to the task
                    # vars
                    var_ids += [sf.variables_taskvar_id, sf.variables_hostvars_id]

                for vid in var_ids:
                    ansible_var = self.variables[vid]
                    ansible_var.set(repl_command.name, new_value)

            else:
                # Error during parsing or --help was requested
                value = str(repl_command)

        elif request.context in ["repl", "watch", "clipboard", "variables"] and request.frame_id:
            sf = self.stackframes[request.frame_id]
            value, value_type = self._safe_evaluate_expression(expression, sf.task_vars)

        else:
            value = f"Evaluation for {request.context} is not implemented"

        return ansibug.dap.EvaluateResponse(
            request_seq=request.seq,
            result=value,
            type=value_type,
        )

    def get_stacktrace(
        self,
        request: ansibug.dap.StackTraceRequest,
    ) -> ansibug.dap.StackTraceResponse:
        stack_frames: list[ansibug.dap.StackFrame] = []
        thread = self.threads[request.thread_id]
        for sfid in thread.stack_frames:
            sf = self.stackframes[sfid]
            stack_frames.append(sf.to_dap())

        return ansibug.dap.StackTraceResponse(
            request_seq=request.seq,
            stack_frames=stack_frames,
            total_frames=len(stack_frames),
        )

    def get_threads(
        self,
        request: ansibug.dap.ThreadsRequest,
    ) -> ansibug.dap.ThreadsResponse:
        return ansibug.dap.ThreadsResponse(
            request_seq=request.seq,
            threads=[t.to_dap() for t in self.threads.values()],
        )

    def get_variables(
        self,
        request: ansibug.dap.VariablesRequest,
    ) -> ansibug.dap.VariablesResponse:
        variable = self.variables[request.variables_reference]

        variables: list[ansibug.dap.Variable] = []
        for name, value in variable.get():
            child_var: AnsibleVariable | None = None
            if isinstance(value, (collections.abc.Mapping, collections.abc.Sequence)) and not isinstance(value, str):
                child_var = self._add_variable(variable.stackframe, value)

            variables.append(
                ansibug.dap.Variable(
                    name=name,
                    value=repr(value),
                    type=type(value).__name__,
                    named_variables=child_var.named_variables if child_var else 0,
                    indexed_variables=child_var.indexed_variables if child_var else 0,
                    variables_reference=child_var.id if child_var else 0,
                )
            )

        return ansibug.dap.VariablesResponse(
            request_seq=request.seq,
            variables=variables,
        )

    def set_variable(
        self,
        request: ansibug.dap.SetVariableRequest,
    ) -> ansibug.dap.SetVariableResponse:
        variable = self.variables[request.variables_reference]

        new_value = self._template(request.value, variable.stackframe.task_vars)
        variable.set(request.name, new_value)

        new_container = None
        if isinstance(new_value, (collections.abc.Mapping, collections.abc.Sequence)) and not isinstance(
            new_value, str
        ):
            new_container = self._add_variable(variable.stackframe, new_value)

        return ansibug.dap.SetVariableResponse(
            request_seq=request.seq,
            value=repr(new_value),
            type=type(new_value).__name__,
            variables_reference=new_container.id if new_container else 0,
            named_variables=new_container.named_variables if new_container else 0,
            indexed_variables=new_container.indexed_variables if new_container else 0,
        )

    def step_in(
        self,
        request: ansibug.dap.StepInRequest,
    ) -> None:
        self._resume_threads([request.thread_id], ThreadState.STEP_IN)

    def step_out(
        self,
        request: ansibug.dap.StepOutRequest,
    ) -> None:
        self._resume_threads([request.thread_id], ThreadState.STEP_OUT)

    def step_over(
        self,
        request: ansibug.dap.NextRequest,
    ) -> None:
        self._resume_threads([request.thread_id], ThreadState.STEP_OVER)

    def _add_thread(
        self,
        host: Host,
    ) -> AnsibleThread:
        tid = self._debugger.next_thread_id()

        thread = self.threads[tid] = AnsibleThread(
            id=tid,
            host=host,
        )
        self._debugger.queue_msg(
            ansibug.dap.ThreadEvent(
                reason="started",
                thread_id=tid,
            )
        )

        return thread

    def _add_variable(
        self,
        stackframe: AnsibleStackFrame,
        value: collections.abc.Mapping[t.Any, t.Any] | collections.abc.Sequence[t.Any],
        var_factory: t.Callable[[int], AnsibleVariable] | None = None,
    ) -> AnsibleVariable:
        var_id = self._debugger.next_variable_id()

        if var_factory:
            var = var_factory(var_id)

        elif isinstance(value, collections.abc.Mapping):
            var = AnsibleDictVariable(var_id, stackframe, value)

        else:
            var = AnsibleListVariable(var_id, stackframe, value)

        self.variables[var_id] = var
        stackframe.variables.append(var_id)

        return var

    def _resume_threads(
        self,
        thread_ids: collections.abc.Iterable[int],
        thread_state: ThreadState,
    ) -> None:
        with self._waiting_condition:
            for tid in thread_ids:
                self.threads[tid].set_state(thread_state)

            self._waiting_condition.notify_all()

    def _safe_evaluate_expression(
        self,
        expression: str,
        task_vars: dict[t.Any, t.Any],
    ) -> tuple[t.Any, str | None]:
        """Evaluates an expression with a fallback on exception."""
        value_type = None
        try:
            templated_value = self._template(expression, task_vars)
        except AnsibleUndefinedVariable as e:
            value = f"{type(e).__name__}: {e!s}"
        except Exception as e:
            value = traceback.format_exc()
        else:
            value = repr(templated_value)
            value_type = type(templated_value).__name__

        return value, value_type

    def _template(
        self,
        value: str,
        variables: dict[t.Any, t.Any],
    ) -> t.Any:
        templar = Templar(loader=self._loader, variables=variables)

        # Always use native types even if the config has not been enabled as it
        # allows expressions like `1` to be returned as an int and keeps things
        # consistent.
        if not C.DEFAULT_JINJA2_NATIVE:
            templar = templar.copy_with_new_env(environment_class=AnsibleNativeEnvironment)

        expression = "{{ %s }}" % value

        return templar.template(expression)


class StrategyModule(LinearStrategy):
    def __init__(
        self,
        tqm: TaskQueueManager,
    ) -> None:
        super().__init__(tqm)

        # Used for type annotation checks, technically defined in __init__ as well
        self._tqm = tqm

        # Set in run()
        self._debug_state: AnsibleDebugState

    def _execute_meta(
        self,
        task: Task,
        play_context: PlayContext,
        iterator: PlayIterator,
        target_host: Host,
    ) -> list[dict[str, t.Any]]:
        """Called when a meta task is about to run"""
        # This is super ick but recreating the vars just for meta tasks would
        # involve too many internal attributes so just get it from the parent
        # locals.

        task_vars = inspect.currentframe().f_back.f_locals["task_vars"]  # type: ignore[union-attr]  # I know this is bad
        self._debug_state.start_task(target_host, task, task_vars)
        res = super()._execute_meta(task, play_context, iterator, target_host)

        meta_action = task.args.get("_raw_params", None)
        if meta_action == "refresh_inventory":
            # refresh_inventory will update the uuid on the host object, we
            # need to mark the existing threads as exited so the next run will
            # pick up the correct changes. It will automatically start those
            # threads as they run.
            self._debug_state.exit_threads()

        else:
            # Needed to ensure the stackframe is removed as these meta tasks
            # have no results for _process_pending_results to handle.
            self._debug_state.end_task(target_host, task)

        return res

    def _load_included_file(
        self,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> list[Block]:
        included_blocks = super()._load_included_file(*args, **kwargs)

        # Need to register these blocks as valid breakpoints and update the client bps
        register_block_breakpoints(self._debug_state._debugger, included_blocks)

        return included_blocks

    def _queue_task(
        self,
        host: Host,
        task: Task,
        task_vars: dict[str, t.Any],
        play_context: PlayContext,
    ) -> None:
        """Called just as a task is about to be queue"""
        self._debug_state.start_task(host, task, task_vars)

        return super()._queue_task(host, task, task_vars, play_context)

    def _process_pending_results(
        self,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> list[TaskResult]:
        """Called when gathering the results of a queued task."""
        res = super()._process_pending_results(*args, **kwargs)

        for task_res in res:
            self._debug_state.end_task(task_res._host, task_res._task)

        return res

    def run(
        self,
        iterator: PlayIterator,
        play_context: PlayContext,
    ) -> int:
        """Main strategy entrypoint.

        This is the main strategy entrypoint that is called per play. The first
        step is to associate the current strategy with the debuggee adapter so
        it can respond to breakpoint and other information.
        """
        debugger = AnsibleDebugger()
        self._debug_state = AnsibleDebugState(
            debugger,
            self._loader,
            self._variable_manager,
        )
        try:
            with debugger.with_strategy(self._debug_state):
                try:
                    return super().run(iterator, play_context)
                finally:
                    self._debug_state.exit_threads()

        except EndStrategy:
            display.display("Debugger has requested the process to terminate")
            return TaskQueueManager.RUN_FAILED_BREAK_PLAY



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from ._adapter import DebugAdapterConnection
from ._events import (
    BreakpointEvent,
    ExitedEvent,
    InitializedEvent,
    OutputEvent,
    StoppedEvent,
    StoppedReason,
    TerminatedEvent,
    ThreadEvent,
)
from ._messages import (
    Command,
    Event,
    EventType,
    MessageType,
    ProtocolMessage,
    Request,
    Response,
)
from ._requests import (
    AttachRequest,
    CancelRequest,
    ConfigurationDoneRequest,
    ContinueRequest,
    DisconnectRequest,
    EvaluateRequest,
    InitializeRequest,
    LaunchRequest,
    NextRequest,
    RunInTerminalRequest,
    ScopesRequest,
    SetBreakpointsRequest,
    SetExceptionBreakpointsRequest,
    SetVariableRequest,
    StackTraceRequest,
    StepInRequest,
    StepOutRequest,
    TerminateRequest,
    ThreadsRequest,
    VariablesRequest,
)
from ._responses import (
    AttachResponse,
    CancelResponse,
    ConfigurationDoneResponse,
    ContinueResponse,
    DisconnectResponse,
    ErrorResponse,
    EvaluateResponse,
    InitializeResponse,
    LaunchResponse,
    NextResponse,
    RunInTerminalResponse,
    ScopesResponse,
    SetBreakpointsResponse,
    SetExceptionBreakpointsResponse,
    SetVariableResponse,
    StackTraceResponse,
    StepInResponse,
    StepOutResponse,
    TerminateResponse,
    ThreadsResponse,
    VariablesResponse,
)
from ._types import (
    Breakpoint,
    Capabilities,
    Checksum,
    ExceptionFilterOptions,
    ExceptionOptions,
    ExceptionPathSegment,
    Message,
    Scope,
    Source,
    SourceBreakpoint,
    StackFrame,
    StackFrameFormat,
    Thread,
    ValueFormat,
    Variable,
    VariablePresentationHint,
)

__all__ = [
    "AttachRequest",
    "AttachResponse",
    "Breakpoint",
    "BreakpointEvent",
    "CancelRequest",
    "CancelResponse",
    "Capabilities",
    "Checksum",
    "Command",
    "ConfigurationDoneRequest",
    "ConfigurationDoneResponse",
    "ContinueRequest",
    "ContinueResponse",
    "DebugAdapterConnection",
    "DisconnectRequest",
    "DisconnectResponse",
    "ErrorResponse",
    "EvaluateRequest",
    "EvaluateResponse",
    "Event",
    "EventType",
    "ExceptionFilterOptions",
    "ExceptionOptions",
    "ExceptionPathSegment",
    "ExitedEvent",
    "InitializedEvent",
    "InitializeRequest",
    "InitializeResponse",
    "LaunchRequest",
    "LaunchResponse",
    "Message",
    "MessageType",
    "NextRequest",
    "NextResponse",
    "OutputEvent",
    "ProtocolMessage",
    "Request",
    "Response",
    "RunInTerminalRequest",
    "RunInTerminalResponse",
    "Scope",
    "ScopesRequest",
    "ScopesResponse",
    "SetBreakpointsRequest",
    "SetBreakpointsResponse",
    "SetExceptionBreakpointsRequest",
    "SetExceptionBreakpointsResponse",
    "SetVariableRequest",
    "SetVariableResponse",
    "Source",
    "SourceBreakpoint",
    "StackFrame",
    "StackFrameFormat",
    "StackTraceRequest",
    "StackTraceResponse",
    "StepInRequest",
    "StepInResponse",
    "StepOutRequest",
    "StepOutResponse",
    "StoppedEvent",
    "StoppedReason",
    "TerminatedEvent",
    "TerminateRequest",
    "TerminateResponse",
    "Thread",
    "ThreadEvent",
    "ThreadsRequest",
    "ThreadsResponse",
    "ValueFormat",
    "Variable",
    "VariablePresentationHint",
    "VariablesRequest",
    "VariablesResponse",
]
